@@ -1,3 +1,5 @@
+#!/usr/bin/env perl
+
 use TAP::Harness;
 
 my ($skip_qregex) = @ARGV;
@@ -14,11 +16,11 @@ my @serialization = glob "t/serialization/{01,02,03}*.t";
 my @qast = glob "t/qast/02*";
 my @js_specific = "t/js/getcomp-js.t";
 
-my @nqp_tests = grep {$_ !~ /regex-interpolation/} glob "t/nqp/*.t";
+my @nqp_tests = grep {$_ !~ /regex-interpolation/ && $_ !~ /optional-args/ && $_ !~ /112|106/} glob "t/nqp/*.t";
 
 my @moar_tests = (); # qw(t/moar/03-line-seps.t);
 
-my @runtime_unit_tests = qw(t/js/varint.js);
+my @runtime_unit_tests; # = qw(t/js/varint.js);
 my @continuation_tests = qw(t/js/continuations.t);
 
 my @native_call_tests = qw(t/nativecall/01-basic.t);
