@@ -1,10 +1,16 @@
 'use strict';
-var NQPExceptionWithCtx = require('./nqp-exception-with-ctx.js');
-var NQPObject = require('./nqp-object.js');
-var Null = require('./null.js');
-var exceptionsStack = require('./exceptions-stack.js');
+const NQPExceptionWithCtx = require('./nqp-exception-with-ctx.js');
+const NQPException = require('./nqp-exception.js');
+const NQPObject = require('./nqp-object.js');
+const Null = require('./null.js');
 
-var BOOT = require('./BOOT.js');
+const BOOT = require('./BOOT.js');
+
+const NQPInt = require('./nqp-int.js');
+const NQPStr = require('./nqp-str.js');
+
+
+const stackTrace = require('stack-trace');
 
 const NEXT = 4;
 const REDO = 8;
@@ -19,7 +25,7 @@ const AWAIT = 8192;
 const EMIT = 16384;
 const DONE = 32768;
 
-let categoryIDs = {
+const categoryIDs = {
   NEXT: NEXT,
   REDO: REDO,
   LAST: LAST,
@@ -33,8 +39,8 @@ let categoryIDs = {
   DONE: DONE,
 };
 
-let categoryToName = {};
-for (let name of Object.keys(categoryIDs)) {
+const categoryToName = {};
+for (const name of Object.keys(categoryIDs)) {
   categoryToName[categoryIDs[name]] = name;
 }
 
@@ -57,61 +63,67 @@ class Ctx extends NQPObject {
   }
 
   last() {
-    this.controlException(LAST);
+    return this.controlException(LAST);
   }
 
   lastLabeled(label) {
-    this.controlExceptionLabeled(label, LAST);
+    return this.controlExceptionLabeled(label, LAST);
   }
 
   next() {
-    this.controlException(NEXT);
+    return this.controlException(NEXT);
   }
 
   nextLabeled(label) {
-    this.controlExceptionLabeled(label, NEXT);
+    return this.controlExceptionLabeled(label, NEXT);
   }
 
   redo() {
-    this.controlException(REDO);
+    return this.controlException(REDO);
   }
 
   redoLabeled(label) {
-    this.controlExceptionLabeled(label, REDO);
+    return this.controlExceptionLabeled(label, REDO);
   }
 
   controlException(category) {
-    let exType = BOOT.Exception;
-    let exception = exType._STable.REPR.allocate(exType._STable);
+    const exType = BOOT.Exception;
+    const exception = exType.$$STable.REPR.allocate(exType.$$STable);
     exception.$$category = category;
-    this.propagateControlException(exception);
+    return this.propagateControlException(exception);
   }
 
   controlExceptionLabeled(label, category) {
-    let exType = BOOT.Exception;
-    let exception = exType._STable.REPR.allocate(exType._STable);
+    const exType = BOOT.Exception;
+    const exception = exType.$$STable.REPR.allocate(exType.$$STable);
     exception.$$category = category | LABELED;
     exception.$$payload = label;
-    this.propagateControlException(exception);
+    return this.propagateControlException(exception);
   }
 
-  propagateControlException(exception) {
-    let handler = '$$' + categoryToName[exception.$$category & ~LABELED];
-    let labeled = exception.$$category & LABELED;
+  /*async*/ propagateControlException(exception) {
+    const handler = '$$' + categoryToName[exception.$$category & ~LABELED];
+    const labeled = exception.$$category & LABELED;
 
-    var ctx = this;
+    let ctx = this;
 
     while (ctx) {
-      if ((ctx[handler] || ctx.$$CONTROL) && (!labeled || ctx.$$label === exception.$$payload)) {
+      if (ctx.$$controlHandlerOuter) {
+        ctx = ctx.$$controlHandlerOuter;
+      }
+
+      if (ctx.$$CONTROL || (ctx[handler] && (!labeled || ctx.$$label === exception.$$payload))) {
         exception.caught = ctx;
         ctx.exception = exception;
 
-        exceptionsStack.push(exception);
         try {
+          const wrapped = new Ctx(this, this, null);
+          wrapped.$$currentException = exception;
+
           if (ctx[handler]) {
-            ctx.unwind.ret = ctx[handler]();
+            ctx.$$unwind.ret = /*await*/ ctx[handler](wrapped);
           } else {
-            ctx.unwind.ret = ctx.$$CONTROL();
+            ctx.$$unwind.ret = /*await*/ ctx.$$CONTROL(wrapped);
           }
         } catch (e) {
           if (e instanceof ResumeException && e.exception === exception) {
@@ -119,67 +131,93 @@ class Ctx extends NQPObject {
           } else {
             throw e;
           }
-        } finally {
-          exceptionsStack.pop();
         }
 
-        throw ctx.unwind;
+        throw ctx.$$unwind;
       }
       ctx = ctx.$$caller;
     }
 
-    throw exception;
+    const uncaught_control = this.$$getHLL().get('uncaught_control');
+    if (uncaught_control) {
+      try {
+        /*await*/ uncaught_control.$$call(this, null, exception.$$decont(this));
+      } catch (e) {
+        if (e instanceof ResumeException && e.exception === exception) {
+          return;
+        } else {
+          throw e;
+        }
+      }
+    } else {
+      // TODO maybe wrap this up?
+      throw exception;
+    }
   }
 
-  propagateException(exception) {
+  /*async*/ propagateException(exception) {
     if (exception.$$category) {
-      this.propagateControlException(exception);
-      return;
+      return this.propagateControlException(exception);
     }
 
-    var ctx = this;
+    let ctx = this;
 
     while (ctx) {
+      if (ctx.$$catchHandlerOuter) {
+        ctx = ctx.$$catchHandlerOuter;
+      }
+
       if (ctx.$$CATCH) {
         exception.caught = ctx;
         ctx.exception = exception;
 
-        exceptionsStack.push(exception);
+
         try {
-          ctx.unwind.ret = ctx.$$CATCH();
+          const wrapped = new Ctx(this, this, null);
+          wrapped.$$currentException = exception;
+          ctx.$$unwind.ret = /*await*/ ctx.$$CATCH(wrapped);
         } catch (e) {
           if (e instanceof ResumeException && e.exception === exception) {
             return;
           } else {
             throw e;
           }
-        } finally {
-          exceptionsStack.pop();
         }
-
-        throw ctx.unwind;
+        throw ctx.$$unwind;
       }
       ctx = ctx.$$caller;
     }
-    throw exception;
-  }
 
-  catchException(exception) {
-    this.exception = exception;
-    exceptionsStack.push(exception);
-    try {
-      return this.$$CATCH();
-    } finally {
-      exceptionsStack.pop();
+    const uncaught_exception = this.$$getHLL().get('uncaught_exception');
+    if (uncaught_exception) {
+      try {
+        /*await*/ uncaught_exception.$$call(this, null, exception);
+      } catch (e) {
+        if (e instanceof ResumeException && e.exception === exception) {
+          return;
+        } else {
+          throw e;
+        }
+      }
+    } else {
+      // TODO maybe wrap this up?
+      throw exception;
     }
   }
 
+  /*async*/ catchException(exception) {
+    // we don't have access to the most correct ctx in case of this sort of exception
+    const wrapped = new Ctx(this, this, null);
+    wrapped.$$currentException = exception;
+    return /*await*/ this.$$CATCH(wrapped);
+  }
+
   rethrow(exception) {
-    this.propagateException(exception);
+    return this.propagateException(exception);
   }
 
   die(msg) {
-    this.propagateException(new NQPExceptionWithCtx(msg, this));
+    return this.propagateException(new NQPExceptionWithCtx(msg, this, stackTrace.get()));
   }
 
   resume(exception) {
@@ -187,43 +225,57 @@ class Ctx extends NQPObject {
   }
 
   throw(exception) {
-    this.propagateException(exception);
+    // Partial workaround for browser that don't use the V8 stack trace API (like Firefox)
+    try {
+      exception.$$stack = stackTrace.get();
+    } catch (e) {
+      exception.$$stack = [];
+    }
+    exception.$$ctx = this;
+    return this.propagateException(exception);
   }
 
   throwpayloadlexcaller(category, payload) {
     let ctx = this.$$skipHandlers().$$caller;
-    let isThunkOrCompilerStub = code => code.staticCode.isThunk || code.isCompilerStub;
+    const isThunkOrCompilerStub = code => code !== null && (code.staticCode.isThunk || code.isCompilerStub);
     while (ctx && isThunkOrCompilerStub(ctx.codeRef())) {
       ctx = ctx.$$caller;
     }
 
-    this.$$throwLexicalException(ctx, category, payload);
+    return this.$$throwLexicalException(ctx, category, payload);
   }
 
   throwpayloadlex(category, payload) {
-    this.$$throwLexicalException(this, category, payload);
+    return this.$$throwLexicalException(this, category, payload);
   }
 
-  $$throwLexicalException(lookFrom, category, payload) {
-    let exType = BOOT.Exception;
-    let exception = exType._STable.REPR.allocate(exType._STable);
+  /*async*/ $$throwLexicalException(lookFrom, category, payload) {
+    const exType = BOOT.Exception;
+    const exception = exType.$$STable.REPR.allocate(exType.$$STable);
+
     exception.$$category = category;
     exception.$$payload = payload;
-    let handler = '$$' + categoryToName[category];
+    const handler = '$$' + categoryToName[category];
 
     let ctx = lookFrom;
 
     while (ctx) {
+      if (ctx.$$controlHandlerOuter) {
+        ctx = ctx.$$controlHandlerOuter;
+      }
+
+      // TODO - think about checking the label
       if (ctx[handler] || ctx.$$CONTROL) {
         exception.caught = ctx;
         ctx.exception = exception;
 
-        exceptionsStack.push(exception);
         try {
+          const wrapped = new Ctx(this, this, null);
+          wrapped.$$currentException = exception;
           if (ctx[handler]) {
-            ctx.unwind.ret = ctx[handler]();
+            ctx.$$unwind.ret = /*await*/ ctx[handler](wrapped);
           } else {
-            ctx.unwind.ret = ctx.$$CONTROL();
+            ctx.$$unwind.ret = /*await*/ ctx.$$CONTROL(wrapped);
           }
         } catch (e) {
           if (e instanceof ResumeException && e.exception === exception) {
@@ -231,20 +283,28 @@ class Ctx extends NQPObject {
           } else {
             throw e;
           }
-        } finally {
-          exceptionsStack.pop();
         }
 
-        throw ctx.unwind;
+        let check = lookFrom;
+        while (check) {
+          if (check === ctx) {
+            throw ctx.$$unwind;
+          }
+          check = check.$$caller;
+        }
+
+        /*await*/ this.$$getHLL().get('lexical_handler_not_found_error').$$call(this, null, new NQPInt(category), new NQPInt(1));
+        return;
       }
+
       ctx = ctx.$$outer;
     }
 
-    this.$$getHLL().get('lexical_handler_not_found_error').$$call(this, null, category, 0);
+    /*await*/ this.$$getHLL().get('lexical_handler_not_found_error').$$call(this, null, new NQPInt(category), new NQPInt(0));
   }
 
   lookupDynamic(name) {
-    var ctx = this;
+    let ctx = this;
     while (ctx) {
       if (ctx.hasOwnProperty(name)) {
         return ctx[name];
@@ -256,8 +316,16 @@ class Ctx extends NQPObject {
        nqp code usually fallbacks to looking up of global */
   }
 
+  lookupOrNull(name) {
+    if (this.hasOwnProperty(name)) {
+      return this[name];
+    } else {
+      return Null;
+    }
+  }
+
   lookupDynamicFromCaller(name) {
-    var ctx = this.$$caller;
+    let ctx = this.$$caller;
     while (ctx) {
       if (ctx.hasOwnProperty(name)) {
         return ctx[name];
@@ -270,9 +338,9 @@ class Ctx extends NQPObject {
   }
 
   lookupWithCallers(name) {
-    var currentCallerCtx = this;
+    let currentCallerCtx = this;
     while (currentCallerCtx) {
-      var currentCtx = currentCallerCtx;
+      let currentCtx = currentCallerCtx;
       while (currentCtx) {
         if (currentCtx.hasOwnProperty(name)) {
           return currentCtx[name];
@@ -285,7 +353,7 @@ class Ctx extends NQPObject {
   }
 
   lookup(name) {
-    var ctx = this;
+    let ctx = this;
     while (ctx) {
       if (ctx.hasOwnProperty(name)) {
         return ctx[name];
@@ -300,8 +368,24 @@ class Ctx extends NQPObject {
     return this.$$outer.lookup(name);
   }
 
+  $$exception() {
+    let ctx = this;
+    while (ctx) {
+      if (ctx.$$currentException) {
+        return ctx.$$currentException;
+      }
+      ctx = ctx.$$caller;
+    }
+
+    console.log('exception not found');
+  }
+
+  $$lastexpayload() {
+    return this.$$exception().$$payload;
+  }
+
   $$getHLL() {
-    var ctx = this;
+    let ctx = this;
     while (ctx) {
       if (ctx.$$hll) {
         return ctx.$$hll;
@@ -310,8 +394,15 @@ class Ctx extends NQPObject {
     }
   }
 
-  $$atkey(key) {
-    return this.lookup(key);
+  $$atkey(name) {
+    let ctx = this;
+    while (ctx) {
+      if (ctx.hasOwnProperty(name)) {
+        return ctx[name];
+      }
+      ctx = ctx.$$outer;
+    }
+    throw new NQPException(`Lexical with name '${name}' does not exist in this frame`);
   }
 
   $$bindkey(key, value) {
@@ -328,7 +419,7 @@ class Ctx extends NQPObject {
   }
 
   bind(name, value) {
-    var ctx = this;
+    let ctx = this;
     while (ctx) {
       if (ctx.hasOwnProperty(name)) {
         ctx[name] = value;
@@ -340,7 +431,19 @@ class Ctx extends NQPObject {
   }
 
   bindDynamic(name, value) {
-    var ctx = this;
+    let ctx = this;
+    while (ctx) {
+      if (ctx.hasOwnProperty(name)) {
+        ctx[name] = value;
+        return ctx[name];
+      }
+      ctx = ctx.$$caller;
+    }
+    throw `Can't bind dynamic: ${name}`;
+  }
+
+  bindDynamicFromCaller(name, value) {
+    let ctx = this.$$caller;
     while (ctx) {
       if (ctx.hasOwnProperty(name)) {
         ctx[name] = value;
@@ -390,15 +493,15 @@ class CtxIter extends NQPObject {
     return this.$$idx < this.$$target;
   }
 
-  Str(ctx, _NAMED, self) {
-    return this.$$iterkey_s();
+  p6$Str(ctx, _NAMED, self) {
+    return new NQPStr(this.$$iterkey_s());
   }
 
-  key(ctx, _NAMED, self) {
-    return this.$$iterkey_s();
+  p6$key(ctx, _NAMED, self) {
+    return new NQPStr(this.$$iterkey_s());
   }
 
-  value(ctx, _NAMED, self) {
+  p6$value(ctx, _NAMED, self) {
     return this.$$iterval();
   }
 };
