@@ -1,9 +1,9 @@
-let NQPException = require('./nqp-exception');
+const NQPException = require('./nqp-exception.js');
 class NQPExceptionWithCtx extends NQPException {
-  constructor(message, ctx) {
+  constructor(message, ctx, stack) {
     super(message);
     this.$$ctx = ctx;
+    this.$$stack = stack;
   }
-
 };
 module.exports = NQPExceptionWithCtx;
