@@ -1,11 +1,13 @@
+const NQPStr = require('./nqp-str.js');
+
 class NQPException extends Error {
   constructor(message) {
     super(message);
     this.$$message = message;
   }
 
-  Str(ctx, _NAMED, self) {
-    return this.$$message;
+  p6$Str(ctx, _NAMED, self) {
+    return new NQPStr(this.$$message);
   }
 
   $$toBool(ctx) {
@@ -25,10 +27,10 @@ class NQPException extends Error {
   }
 };
 
-let proto = NQPException.prototype; /* Avoid gjslint warning */
-proto._STable = {
+const proto = NQPException.prototype; /* Avoid gjslint warning */
+proto.$$STable = {
   HOW: {
-    name: function(ctx, _NAMED, how, obj) {
+    p6$name: function(ctx, _NAMED, how, obj) {
       return 'BOOTException';
     },
     $$decont(ctx) {
