@@ -74,6 +74,8 @@ sub configure_refine_vars {
               || File::Spec->catdir( $config->{'prefix'}, 'share', 'nqp' )
         )
     );
+
+    $config->{static_nqp_home} = $config->{relocatable} eq 'reloc' ? '' : $config->{nqp_home};
 }
 
 sub configure_misc {
@@ -115,7 +117,7 @@ sub configure_moar_backend {
     }
     else {
         my $qchar = $config->{quote};
-        $imoar->{config}{static_nqp_home} = $config->{nqp_home};
+        $imoar->{config}{static_nqp_home} = $config->{static_nqp_home};
         $imoar->{config}{static_nqp_home_define} =
             '-DSTATIC_NQP_HOME='
           . $qchar
@@ -216,7 +218,7 @@ sub configure_jvm_backend {
         print "got: $_";
         if (/(?:java|jdk) version "(\d+)(?:\.(\d+))?/) {
             $jvm_found = 1;
-            if ( $1 > 1 || $1 == 1 && $2 >= 8 ) {
+            if ( $1 > 1 || $1 == 1 && $2 >= 9 ) {
                 $ijvm->{ok} = $jvm_ok = 1;
             }
             $got = $_;
@@ -228,7 +230,7 @@ sub configure_jvm_backend {
         push @errors, "No JVM (java executable) in path; cannot continue";
     }
     elsif ( !$jvm_ok ) {
-        push @errors, "Need at least JVM 1.8 (got $got)";
+        push @errors, "Need at least JVM 1.9 (got $got)";
     }
     $self->sorry(@errors) if @errors;
 
@@ -244,8 +246,11 @@ sub configure_jars {
         {
             asm => [qw<3rdparty asm asm-4.1.jar>],
             'asm-tree' => [qw<3rdparty asm asm-tree-4.1.jar>],
+            fastutil => [qw<3rdparty fastutil fastutil-8.5.13-min.jar>],
             jline => [qw<3rdparty jline jline-1.0.jar>],
-            jna => [qw<3rdparty jna jna-4.0.0.jar>],
+            jna => [qw<3rdparty jna jna-4.5.0.jar>],
+            'jna-platform' => [qw<3rdparty jna jna-platform-4.5.0.jar>],
+            lz4 => [qw<3rdparty lz4 lz4-java-1.8.0.jar>],
         }
     );
 }
@@ -327,9 +332,13 @@ sub gen_moar {
     my $gen_moar      = $options->{'gen-moar'};
     my $force_rebuild = $options->{'force-rebuild'};
     my $has_gen_moar  = defined $gen_moar;
+
     my @opts          = @{ $options->{'moar-option'} || [] };
-    push @opts, "--optimize";
+    push @opts, "--optimize" unless @opts && grep { $_ eq "--no-optimize" } @opts;
     push @opts, '--relocatable' if $options->{relocatable};
+    push @opts,
+        '--git-cache-dir=' . File::Spec->rel2abs($options->{'git-cache-dir'})
+        if $options->{'git-cache-dir'};
     my $startdir     = $config->{base_dir};
     my $git_protocol = $options->{'git-protocol'} || 'https';
     my $try_generate;
